#include "Lexer.h"
#include "Parser.h"
#include "Optimization.h"
#include <stdio.h>
#include <map>
#include "llvm/Transforms/Scalar.h"
#include "llvm/Transforms/Scalar/GVN.h"
#include "KaleidoscopeJIT.h"
#include "llvm/Support/TargetSelect.h"
#include "llvm/Target/TargetMachine.h"
void HandleDefinition();
void HandleExtern();
void HandleTopLevelExpression();

using namespace llvm;
using namespace llvm::orc;
extern std::map<char, int> BinOpPrecedence;
extern char CurTok;
extern std::unique_ptr<Module> TheModule;
extern LLVMContext TheContext;
extern std::map<std::string, std::unique_ptr<PrototypeAST>> FunctionProtos;
std::unique_ptr<KaleidoscopeJIT> TheJIT;

int main(int argc, char *argv[]){
    BinOpPrecedence['<'] = 10;
    BinOpPrecedence['>'] = 10;
    BinOpPrecedence['+'] = 20;
    BinOpPrecedence['-'] = 20;
    BinOpPrecedence['*'] = 40;

    fprintf(stderr, "ready> ");
    getNextToken();

    InitializeNativeTarget();
    InitializeNativeTargetAsmPrinter();
    InitializeNativeTargetAsmParser();
    TheJIT = llvm::make_unique<KaleidoscopeJIT>();

    TheModule = llvm::make_unique<Module>("my cool jit", TheContext);
    InitializeModuleAndPassManager();
    while(1){
        fprintf(stderr, "ready> ");
        switch (CurTok) {
            case tok_eof:
                return 0;
            case ';':
                getNextToken();
                break;
            case tok_def:
                HandleDefinition();
                break;
            case tok_extern:
                HandleExtern();
                break;
            default:
                HandleTopLevelExpression();
                break;
        }
    }
    TheModule->dump();
    fprintf(stderr, "hi bro");
    return 0;
}

void HandleDefinition() {
    if (auto FnAST = ParseDefinition()) {
        if (auto *FnIR = FnAST->codegen()) {
            fprintf(stderr, "Parsed a function definition.\n");
            FnIR->dump();
            TheJIT->addModule(std::move(TheModule));
            InitializeModuleAndPassManager();
        }
    } else {
        getNextToken();
    }
}

void HandleExtern() {
    if (auto ProtoAST = ParseExtern()) {
        if (auto *FnIR = ProtoAST->codegen()) {
            fprintf(stderr, "Parsed an extern\n");
            FnIR->dump();
            FunctionProtos[ProtoAST->getName()] = std::move(ProtoAST);
        }
    } else {
        getNextToken();
    }
}

void HandleTopLevelExpression() {
    if (auto FnAST = ParseTopLevelExpr()) {
        if (FnAST->codegen()) {
            auto H = TheJIT->addModule(std::move(TheModule));
            InitializeModuleAndPassManager();

            auto ExprSymbol = TheJIT->findSymbol("__anon_expr");
            assert(ExprSymbol && "Function not found");

            double (*FP)() = (double (*)())(intptr_t)ExprSymbol.getAddress();
            fprintf(stderr, "Evaluated to %f\n", FP());

            TheJIT->removeModule(H);
        }
    } else {
        getNextToken();
    }
}

extern "C" double putchard(double X){
    fputc((char)X, stderr);
    return 0;
}
